/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   TrackingSetup.h
/// \date   19.04.2024
/// \brief  A detector setup interface used for tracking input data initialization (header)
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "SubChain.h"
#include "TrackingDefs.h"
#include "tof/TrackingInterface.h"

#include <type_traits>

// TODO: SZh 19.04.2024: Provide interfaces for other subsystems and redefine access

namespace cbm::algo
{
  /// \class TrackingSetup
  /// \brief A detector setup interface class for tracking input data initialization
  class TrackingSetup : public SubChain {
   private:
    /// \brief A helper temporary function to enable a method for a given detector subsystem
    template<ca::EDetectorID DetID>
    static constexpr bool EnableDet()
    {
      return DetID == ca::EDetectorID::Tof;
    }

   public:
    /// \brief Default constructor
    TrackingSetup() = default;

    /// \brief Copy constructor
    TrackingSetup(const TrackingSetup&) = delete;

    /// \brief Move constructor
    TrackingSetup(TrackingSetup&&) = delete;

    /// \brief Destructor
    ~TrackingSetup() = default;

    /// \brief Initializer function
    void Init();

    /// \brief   Returns tracking station index by the TOF address
    /// \param   address  Unique address of a TOF element
    /// \return  Local index of tracking station
    template<ca::EDetectorID DetID, std::enable_if_t<EnableDet<DetID>(), bool> = true>
    int GetTrackingStation(uint32_t address) const
    {
      if constexpr (DetID == ca::EDetectorID::Tof) {
        return fTof.GetTrackingStation(address);
      }
      return -1;
    }


   private:
    tof::TrackingInterface fTof;  ///< TOF tracking interface
  };
}  // namespace cbm::algo
